#!/usr/bin/env python3

from utilities import draw_life_canvas, start_timer
from lifeutils import make_alive, is_alive, is_dead, make_dead, wrap_x, wrap_y
import conway_starters
from conway_starters import *

import random
import drawcanvas
import tkinter

from byugrid import Grid


def random_placement(grid):
    """
    Set 20% of the grid cells to be black
    :param grid: a grid
    :return: the same grid, but with random black cells in it
    >>> grid1 = Grid.build([[None, None, None], [None, None, None], [None, None, None]])
    >>> random.seed(108)
    >>> random_placement(grid1)
    [['b', None, 'b'], ['b', None, None], [None, None, 'b']]
    """
    pass


def count_alive_neighbors(grid, x, y):
    """
    Counts the number of alive neighbors of the cell at (x, y) in the grid.
    There are 8 possible neighbors for all cells, including wrapping when
    we reach the edges.

    :param grid: a grid with some alive cells
    :param x: the x coordinate of a cell
    :param y: the y coordinate of a cell
    :return: the number of alive neighbors of the specified cell
    >>> grid1 = Grid.build([[None, 'b', 'b'], [None, 'b', None], ['b', None, 'b']])
    >>> count_alive_neighbors(grid1, 1, 1)
    4
    >>> count_alive_neighbors(grid1, 0, 1)
    5
    """
    pass


def conway_rules(grid):
    """
    Run Conway's rules of life on the grid and return a new grid that represents
    the next generation. The rules:
    1) Any live cell with two or three live neighbors survives.
    2) Any dead cell with three live neighbors becomes a live cell.
    3) All other cells become or stay dead.
    >>> grid_test = Grid.build([[None]])
    >>> conway_rules(grid_test)
    [[None]]
    >>> grid_test = Grid.build([[None, 'b', None]])
    >>> conway_rules(grid_test)
    [['b', 'b', 'b']]
    >>> grid_test = Grid.build([[None, None, None], [None, 'b', None], [None, None, None]])
    >>> conway_rules(grid_test)
    [[None, None, None], [None, None, None], [None, None, None]]
    >>> block_starter = block()
    >>> conway_rules(block_starter)
    [[None, None, None, None], [None, 'b', 'b', None], [None, 'b', 'b', None], [None, None, None, None]]
    >>> beehive_starter = beehive()
    >>> conway_rules(beehive_starter)
    [[None, None, None, None, None, None], [None, None, 'b', 'b', None, None], [None, 'b', None, None, 'b', None], [None, None, 'b', 'b', None, None], [None, None, None, None, None, None]]
    >>> blinker_starter = blinker()
    >>> conway_rules(blinker_starter)
    [[None, None, None, None, None], [None, None, 'b', None, None], [None, None, 'b', None, None], [None, None, 'b', None, None], [None, None, None, None, None]]
    >>> toad_starter = toad()
    >>> conway_rules(toad_starter)
    [[None, None, None, None, None, None], [None, None, None, 'b', None, None], [None, 'b', None, None, 'b', None], [None, 'b', None, None, 'b', None], [None, None, 'b', None, None, None], [None, None, None, None, None, None]]
    """
    pass

# ************* Utility Functions Below here


def copy_grid(to_grid, from_grid):
    for x in range(to_grid.width):
        for y in range(to_grid.height):
            value = from_grid.get(x, y)
            to_grid.set(x, y, value)
    return to_grid


def movie_action(grid, canvas, scale):
    """Do one round of the move, call in timer."""
    new_grid = conway_rules(grid)
    copy_grid(grid, new_grid)
    draw_life_canvas(grid, canvas, scale)


def main():
    import argparse

    # setup the argument parser
    parser = argparse.ArgumentParser(description="Run Conway's game of life")
    parser.add_argument('--width', type=int, default=30, help='width of the board, 30 by default')
    parser.add_argument('--height', type=int, default=30, help='height of the board, 30 by default')
    parser.add_argument('--starter', default="random", help='starter to use, random by default')
    parser.add_argument('--speed', type=int, default=500, help='speed of each generation, 500ms by default')
    parser.add_argument('--scale', type=int, default=10, help='scale of board, 10 by default')
    parser.add_argument('--offset', type=int, default=0, help='offset to copy in starter, 0 by default')

    # parse the command line arguments
    args = parser.parse_args()

    # initialize the starter
    if args.starter == 'random':
        grid = Grid(args.width, args.height)
        grid = random_placement(grid)
    elif args.starter in dir(conway_starters) and callable(getattr(conway_starters, args.starter)):
        grid = eval(args.starter + '()')
        if args.width <= grid.width:
            args.width = grid.width
        if args.height <= grid.height:
            args.height = grid.height
        starter_grid = grid.copy()
        grid = Grid(args.width, args.height)
        for x in range(starter_grid.width):
            for y in range(starter_grid.height):
                if is_alive(starter_grid, x, y):
                    make_alive(grid, x + args.offset, y + args.offset)

    # draw the canvas
    canvas = drawcanvas.make_canvas(args.width * args.scale, args.height * args.scale, 'Movie')
    draw_life_canvas(grid, canvas, args.scale)

    # start a timer that will call movie_action every 30ms
    start_timer(canvas, args.speed, lambda: movie_action(grid, canvas, args.scale))

    # run the animation
    tkinter.mainloop()


if __name__ == '__main__':
    main()
